var __defProp = Object.defineProperty;
var __getOwnPropDesc = Object.getOwnPropertyDescriptor;
var __getOwnPropNames = Object.getOwnPropertyNames;
var __hasOwnProp = Object.prototype.hasOwnProperty;
var __export = (target, all) => {
  for (var name in all)
    __defProp(target, name, { get: all[name], enumerable: true });
};
var __copyProps = (to, from2, except, desc) => {
  if (from2 && typeof from2 === "object" || typeof from2 === "function") {
    for (let key of __getOwnPropNames(from2))
      if (!__hasOwnProp.call(to, key) && key !== except)
        __defProp(to, key, { get: () => from2[key], enumerable: !(desc = __getOwnPropDesc(from2, key)) || desc.enumerable });
  }
  return to;
};
var __toCommonJS = (mod) => __copyProps(__defProp({}, "__esModule", { value: true }), mod);

// src/index.ts
var src_exports = {};
__export(src_exports, {
  combine: () => combine,
  composeInstrument: () => composeInstrument,
  create: () => create,
  createInstrument: () => createInstrument,
  createProvider: () => createProvider,
  defaultProviderOptions: () => defaultProviderOptions,
  from: () => from,
  setInstrument: () => setInstrument,
  value: () => value
});
module.exports = __toCommonJS(src_exports);

// src/instrument.ts
function isPromise(value2) {
  return value2 === Promise.resolve(value2);
}
var createInstrument = (opts) => {
  return typeof opts === "function" ? opts() : opts;
};
function createInstrumentor(opts) {
  return function(fn, fnName) {
    let name = opts?.name || fn.name || fnName || "anonymous";
    const handler = typeof opts === "function" ? opts() : opts;
    const f = function() {
      let that = this;
      let args = arguments;
      let func = fn;
      try {
        const executeState = handler.onExecute?.({ name, fn, that, args, state: "execute" });
        if (executeState !== void 0) {
          that = executeState.that || that;
          args = executeState.args || args;
          func = executeState.fn || func;
          name = executeState.name || name;
        }
        let result = func.apply(that, args);
        if (isPromise(result)) {
          return result.then((next) => {
            const promiseResultState = handler.onResult?.({ name, fn: func, that, args, state: "result", result: next, isPromise: true });
            if (promiseResultState?.state === "result") {
              return promiseResultState.result;
            }
            if (promiseResultState?.state === "error") {
              throw promiseResultState.error;
            }
            return next;
          }).catch((error) => {
            const errorState = handler.onError?.({ name, fn: func, that, args, error, state: "error", isPromise: true });
            if (errorState?.state === "error") {
              throw errorState.error;
            }
            if (errorState?.state === "result") {
              return errorState.result;
            }
            throw error;
          });
        } else {
          const resultState = handler.onResult?.({ name, fn: func, that, args, state: "result", result: void 0, isPromise: false });
          if (resultState?.state === "error") {
            throw resultState.error;
          }
          if (resultState?.state === "result") {
            return resultState.result;
          }
          return result;
        }
      } catch (error) {
        const errorState = handler.onError?.({ name, fn: func, that, args, error, state: "error", isPromise: false });
        if (errorState?.state === "error") {
          throw errorState.error;
        }
        if (errorState?.state === "result") {
          return errorState.result;
        }
        throw error;
      }
    };
    Object.defineProperty(f, "name", {
      value: name,
      writable: false
    });
    return f;
  };
}
function instrument(container, instrumentFn, maxLevel = 1, currentLevel = 0) {
  if (container === null || container === void 0)
    return container;
  if (currentLevel > maxLevel)
    return container;
  if (
    // there can be more
    typeof container !== "object" || Array.isArray(container)
  ) {
    return container;
  }
  Object.keys(container).forEach((key) => {
    const instance = container[key];
    if (typeof instance === "function") {
      container[key] = instrumentFn(instance, key);
    } else {
      container[key] = instrument(container[key], instrumentFn, maxLevel, currentLevel + 1);
    }
  });
  return container;
}
function nextInstrument(prev, inst) {
  const next = createInstrumentor(createInstrument(inst));
  const mixin = (fn, name) => {
    const instrumented = prev?.(fn, name) || fn;
    return next(instrumented, name);
  };
  return mixin;
}
function composeInstrument(...ins) {
  return ins.reduceRight((prev, next) => {
    return nextInstrument(prev, next);
  }, void 0);
}

// src/index.ts
var defaultProviderOptions = {
  mode: "singleton",
  instrument: createInstrumentor({})
};
function setInstrument(inst) {
  defaultProviderOptions.instrument = nextInstrument(defaultProviderOptions.instrument, inst);
}
function create(provider, dependent, options = defaultProviderOptions) {
  const opts = { ...defaultProviderOptions, ...options };
  const name = opts?.name || provider.name || "anonymous";
  let cached = void 0;
  let dependentRef = dependent;
  async function load() {
    const actualized = dependentRef ? await dependentRef.get() : void 0;
    const provide = await provider(actualized);
    return { provide };
  }
  async function init() {
    if (opts.mode === "prototype") {
      cached = load();
    } else {
      if (cached === void 0) {
        cached = load();
      }
    }
    return cached;
  }
  async function execute(provider2) {
    const { provide } = await init();
    return provider2(provide);
  }
  async function get() {
    const { provide } = await init();
    return provide;
  }
  function prepare(provider2) {
    return async function() {
      const { provide } = await init();
      return provider2.call(this, provide, ...arguments);
    };
  }
  const _inject = async (executor2) => {
    dependentRef = executor2;
  };
  const executor = { execute, get, prepare, _inject };
  Object.defineProperty(executor.get, "name", {
    value: `${name}.get`,
    writable: false
  });
  Object.defineProperty(executor.execute, "name", {
    value: `${name}.execute`,
    writable: false
  });
  Object.defineProperty(executor.prepare, "name", {
    value: `${name}.prepare`,
    writable: false
  });
  if (opts.instrument) {
    instrument(executor, opts.instrument);
  }
  return executor;
}
var createProvider = (provider) => provider;
var value = (value2) => create(() => value2);
var from = (executor) => {
  return {
    provide: (provider, opts = defaultProviderOptions) => create(provider, executor, opts),
    execute: async (executable) => executor.execute(executable),
    prepare: (provider) => executor.prepare(provider)
  };
};
var combine = function(layout) {
  return create(async () => {
    const result = {};
    for (const [key, value2] of Object.entries(layout)) {
      result[key] = await value2.get();
    }
    return result;
  });
};
// Annotate the CommonJS export names for ESM import in node:
0 && (module.exports = {
  combine,
  composeInstrument,
  create,
  createInstrument,
  createProvider,
  defaultProviderOptions,
  from,
  setInstrument,
  value
});
//# sourceMappingURL=index.js.map